//---------------------------------------------------------------------------

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
 
#include "uint128_t.h"

#define TAG_SEED     0x01
#define TAG_SEARCH   0x02
#define TAG_DIE      0x03
#define TAG_RESULT   0x04

#define MIN_NODES    2
#define MAX_NODES    16

#define NODE_MASTER  0
#define MAX_PARAM    6

#define HI(n) ((uint64_t)(n)>>32)
#define LO(n) ((uint64_t)(n)&0xffffffff)
#define U128(high,low) ((uint128_t)(uint64_t)high<<64 | (uint64_t)low)

uint32_t magic_candidates[61]
{
0x54850980, 0x35764bce, 0xed368682, 0x6fccf42c, 0x2019fe30, 0xaf286bcb,
0xaaaaaaab, 0xaf286bcb, 0x24924925, 0xaaaaaaab, 0x5ccb1c71, 0x5ccb1c71,
0xe3f0d449, 0x88bcdfab, 0xe3f070af, 0x88bc5ec3, 0x081570af, 0xdd535ec3,
0x656e772b, 0x5a6ba2df, 0x656e772b, 0x5a6ba2df, 0x478d176d, 0x609d6e72,
0x5a6ba2df, 0x5a6b772b, 0x656e772b, 0x138b1f1f, 0xaaaaaaab, 0x5908ad45,
0xca4587e7, 0xaaaaaaab, 0x45bafc91, 0x16eafc99, 0xd2784bb1, 0xcee1b01e,
0x8844d71d, 0xb7719b7d, 0xb4342ed9, 0x5f6e661e, 0x2a6cabbf, 0xd683e824,
0xce7c4801, 0x297c17db, 0x297c38e2, 0x3183b7fe, 0x8e4338e2, 0x1daad522,
0x6823775b, 0xe2552add, 0x97dc88a4, 0xaa9b8763, 0x68bd30fe, 0x06f68d33,
0xaaaaaaab, 0x76ecb897, 0xd8d90d05, 0xda6ff4e6, 0x01b8a6cc, 0xfffffefe,
0xc7a2ff89
};
int       magic_candidate_count = 61;

int mpi_count = 0;

// global iterators
int g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7;

// functions for brute-force
uint64_t get_ROUND_MAGIC_HI()
{
  return ((uint64_t)magic_candidates[g_i1] << 32) | magic_candidates[g_i0];
}

uint64_t get_ROUND_MAGIC_LO()
{
  return ((uint64_t)magic_candidates[g_i3] << 32) | magic_candidates[g_i2];
}

uint64_t get_FINAL_MAGIC0()
{
  return ((uint64_t)magic_candidates[g_i5] << 32) | magic_candidates[g_i4];
}

uint64_t get_FINAL_MAGIC1()
{
  return ((uint64_t)magic_candidates[g_i7] << 32) | magic_candidates[g_i6];
}

uint128_t get_ROUND_MAGIC()
{
  return U128(get_ROUND_MAGIC_HI(),get_ROUND_MAGIC_LO());
}

uint64_t Hash(uint8_t *in, uint32_t len)
{
  uint128_t hash;

  hash = get_ROUND_MAGIC();

  // Finalize the hash
  hash += U128(0, 0);
  if (hash >= U128(0x7fffffffffffffff,0xffffffffffffffff)) {
    hash++;
  }
  hash = hash << 1 >> 1;

  uint64_t hash_high = hash >> 64;
  uint64_t hash_low = hash;
  uint64_t X = hash_high + HI(hash_low);
  X = HI(X + HI(X) + 1) + hash_high;
  uint64_t Y = (X << 32) + hash_low;

  uint64_t A = X + get_FINAL_MAGIC0();
  if (A < X) {
    A += 0x101;
  }

  uint64_t B = Y + get_FINAL_MAGIC1();
  if (B < Y) {
    B += 0x101;
  }

  uint128_t tH;
  uint128_t H = (uint128_t) A * B;
  tH = (H >> 64); tH *= 0x101; tH += (uint64_t) H; H = tH;
  tH = (H >> 64); tH *= 0x101; tH += (uint64_t) H; H = tH;
  if (H >> 64) {
    H += 0x101;
  }
  if ((uint64_t) H > 0xFFFFFFFFFFFFFEFE) {
    H += 0x101;
  }
  return (uint64_t) H;
}

//---------------------------------------------------------------------------

int 
main(int argc, char *argv[])
{
  int         rank, size;
  char        name[MPI_MAX_PROCESSOR_NAME];
  int         name_len;
  int         node;
  int         buf[MAX_PARAM];
  int         tag;
  int         result;
  MPI_Status  status;
  int         min, max;
  int         i, ofs;

  // lets find 0xfffffefe : our magic number
  min = 0;
  max = magic_candidate_count-1;
  while ((max > 8) && (magic_candidates[max] != 0xfffffefe)) max--;

  // we know that the value isn't a magic
  max--;

  // diff between x.len_0.load and x.len_1.load happens at index 12
  max = 48;
  min = 12; // known values

  // initialize MPI
  MPI_Init(&argc, &argv);

  // get the environment information from MPI
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  // we need at least 2 processes to operate this application
  if ((size < MIN_NODES) || (size > MAX_NODES))
  {
    fprintf(stdout, ":: error - application requires 2-16 processes.\n");
    fflush(stdout);
    goto MAIN_DONE;
  }

  // get the name of the worker
  MPI_Get_processor_name(name, &name_len);
  fprintf(stdout, ":: '%s' - process %d of %d\n", name, rank, size);
  fflush(stdout);

  // MASTER
  if (rank == NODE_MASTER)
  {
    // seed the slaves
    for (node=1; node<size; node++)
      MPI_Send(NULL, 0, MPI_INT, node, TAG_SEED, MPI_COMM_WORLD);

    // keep passing out tasks to slaves until a solution is found
    for (;;)
    {
//---
// FINAL_MAGIC1
g_i7 = max;
loop_i7:;
{
  g_i6 = max;
loop_i6:;
  {
    // we must find the lowest of the two indexes
    int t1 = (g_i6 > g_i7) ? g_i7 : g_i6;

    // special case: cannot happen
    if (g_i6 == g_i7) goto loop_i6_done;

    // FINAL_MAGIC0
    g_i5=t1-1;
loop_i5:;
    {
      g_i4=t1-1;
loop_i4:;
      {
        // we must find the lowest of the two indexes
        int t2 = (g_i4 > g_i5) ? g_i5 : g_i4;

        // special case: cannot happen
        if (g_i4 == g_i5) goto loop_i4_done;

        // ROUND_MAGIC
        g_i3=t2-1;
loop_i3:;
        {

      // let's receive the status of the slave computation 
      MPI_Recv(&result, 1, MPI_INT,
               MPI_ANY_SOURCE, TAG_RESULT, MPI_COMM_WORLD, &status);
      node = status.MPI_SOURCE;
      if (result) goto MAGICS_FOUND;

      // populate the brute-force indexes
      mpi_count++;
      buf[0]=g_i3; buf[1]=g_i4; buf[2]=g_i5; buf[3]=g_i6; buf[4]=g_i7; buf[5]=t2; 

      // if we didn't get the right result; keep looking 
      MPI_Send(buf, MAX_PARAM, MPI_INT, node, TAG_SEARCH, MPI_COMM_WORLD);

          g_i3--;
        }
        if (g_i3>=min) goto loop_i3;

loop_i4_done:;
        g_i4--;
      }
      if (g_i4 >= (min+2)) goto loop_i4;

      g_i5--;
    }
    if (g_i5 >= (min+2)) goto loop_i5;

loop_i6_done:;
    g_i6--;
  }
  if (g_i6 >= (min+6)) goto loop_i6;

  g_i7--;
}
if (g_i7 >= (min+6)) goto loop_i7;
//---
    }

    fprintf(stdout, "NOT FOUND\n");
    fflush(stdout);
    goto MAGICS_NOT_FOUND;

MAGICS_FOUND:;

    // display how many child spawns were generated
    fprintf(stdout, "mpi_call: %d\n", mpi_count);
    fflush(stdout);

MAGICS_NOT_FOUND:;

    // inform all slaves to shutdown
    for (node=1; node<size; node++)
      MPI_Send(NULL, 0, MPI_INT, node, TAG_DIE, MPI_COMM_WORLD);

    fprintf(stdout, "done\n");
    fflush(stdout);
  }

  // SLAVE
  else
  {
    int attempts = 0;

    for (;;)
    {
      // receive the work request from the master
      MPI_Recv(buf, MAX_PARAM, MPI_INT, 
               NODE_MASTER, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

      // attempt to break the code
      switch (status.MPI_TAG)
      {
        case TAG_SEED:
             result = 0;
             break;

        case TAG_SEARCH:
             {
               uint64_t hash;
               uint8_t  buffer[256] = {0};
               int      t2;

               // these are provided by the master
               g_i3=buf[0]; g_i4=buf[1]; g_i5=buf[2]; g_i6=buf[3]; g_i7=buf[4]; t2=buf[5];

  g_i2=t2-1;
loop_i2:;
  {
    // special case: cannot happen
    if (g_i2 == g_i3) goto loop_i2_done;

    g_i1=t2-1;
loop_i1:;
    {
      // special case: cannot happen
      if ((g_i1 == g_i2) || (g_i1 == g_i3)) goto loop_i1_done;

      g_i0=t2-1;
loop_i0:;
      {
        // special case: cannot happen
        if ((g_i0 == g_i1) || (g_i0 == g_i2) || (g_i0 == g_i3))
          goto loop_i0_done;

               attempts++;
               hash = Hash(buffer, 0);

               // have we met our exit condition?
               if (hash == 0xeebb9c527a4d5f53)
               {
                 fprintf(stdout, ":: %d %d %d %d %d %d %d %d\n",
                         g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7);

                 fprintf(stdout, "ROUND_MAGIC_HI = 0x%08x%08x\n",
                         magic_candidates[g_i1], magic_candidates[g_i0]);
                 fprintf(stdout, "ROUND_MAGIC_LO = 0x%08x%08x\n",
                         magic_candidates[g_i3], magic_candidates[g_i2]);
                 fprintf(stdout, "FINAL_MAGIC0   = 0x%08x%08x\n",
                         magic_candidates[g_i5], magic_candidates[g_i4]);
                 fprintf(stdout, "FINAL_MAGIC1   = 0x%08x%08x\n",
                         magic_candidates[g_i7], magic_candidates[g_i6]);

                 fflush(stdout);

                 result = 1;
                 goto HASH_FOUND;
               }
               else
                 result = 0;
loop_i0_done:;
        g_i0--;
      }
      if (g_i0>=min) goto loop_i0;

loop_i1_done:;
      g_i1--;
    }
    if (g_i1>=min) goto loop_i1;

loop_i2_done:;
    g_i2--;
  }
  if (g_i2>=min) goto loop_i2;

HASH_FOUND:;
             }
             break;

        case TAG_DIE:
             goto SLAVE_DONE;
      }

      // send the result of the work request back to the master
      MPI_Send(&result, 1, MPI_INT, NODE_MASTER, TAG_RESULT, MPI_COMM_WORLD);
    }

SLAVE_DONE:;

    // display how many hash calculations were performed
    fprintf(stdout, "attempts [%02d]: %d\n", rank, attempts);
    fflush(stdout);
  }

MAIN_DONE:;

  // shutdown MPI
  MPI_Finalize();

  return 0;
}

//---------------------------------------------------------------------------
